/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
* National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
* Operator of Los Alamos National Laboratory.
* This file is distributed subject to a Software License Agreement found
* in the file LICENSE that is included with this distribution. 
\*************************************************************************/
/* viewImageCA.c */

/* viewImageCA.c,v 1.9 2010/06/17 18:43:14 jba Exp */
/************************DESCRIPTION***********************************
  This file contains routines for channel access.
**********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <inttypes.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>
#include <Xm/Xm.h>

#include "cadef.h"
#include "alarm.h"

#include "viewImageErrDialog.h"
#include "viewImageCA.h"

#define CA_PEND_EVENT_TIMEOUT .001

/* externals */
extern XtAppContext appContext;

/* Struct for file descriptor linked list */
struct FDLIST {
  struct FDLIST *prev;
  XtInputId inpid;
  int fd;
};

/* locals */
static XtIntervalId caTimeoutId = (XtIntervalId) 0;
struct FDLIST *lastFdInList=(struct FDLIST *)0;
int toBeConnectedCount = 0;

/* forward declarations */
static void registerCA(void *dummy, int fd, int opened);
static void viewImageCaConnectionEvent(struct connection_handler_args args);
#if 0
static void viewImageCaValueEvent(struct event_handler_args args);
static void viewImageCaReadAccessEvent(struct access_rights_handler_args args);
static void viewImageCaWriteAccessEvent(struct access_rights_handler_args args);
#endif
static void viewImageCaException(struct exception_handler_args args);


/*********************************************************************
 viewImageCaPendEvent
 *********************************************************************/
void viewImageCaPendEvent(double waitSeconds)
{
    int status;

    status = ca_pend_event(waitSeconds);
    if (status != ECA_NORMAL){
        errDialog("ca_pend_event failed: Return status: %s\n",
        ca_message(status));
    }
}


/*********************************************************************
 viewImageCaIsConnected
 *********************************************************************/
short viewImageCaIsConnected(chid chid)
{
    if (!chid) return 0;
    if (ca_field_type(chid) == TYPENOTCONN) return 0;
    else return 1;
}


/*********************************************************************
 viewImageCaPendIO
 *********************************************************************/
int viewImageCaPendIO(double waitSeconds)
{
    int status;

    status = ca_pend_io(waitSeconds);
    if (status != ECA_NORMAL){
        errDialog("ca_pend_io failed: Return status: %s\n",
        ca_message(status));
    }
    return status;
}


/*********************************************************************
 viewImageCaFlushIO
 *********************************************************************/
void viewImageCaFlushIO()
{
    ca_flush_io();
}


/*********************************************************************
 init ca, add excptn handler, reg fd, add timer proc
 *********************************************************************/
void viewImageCaInit()
{
    int status;

    /* Initialize channel access */
    status = ca_task_initialize();
    if (status != ECA_NORMAL){
        errDialog("ca_task_initialize failed: Return status: %s\n",
        ca_message(status));
    }

    /* Register exception handler */
    status = ca_add_exception_event(viewImageCaException, NULL);
    if (status != ECA_NORMAL){
        errDialog("ca_add_exception_event failed: Return status: %s\n",
        ca_message(status));
    }

    /* Register file descriptor callback */
    status = ca_add_fd_registration(registerCA, NULL);
    if (status != ECA_NORMAL){
        errDialog("ca_add_fd_registration failed: Return status: %s\n",
        ca_message(status));
    }
}


/*********************************************************************
 Callback when there is activity on a CA file descriptor
 *********************************************************************/
void viewImageCaPoll()
{
    ca_poll();
    /*ca_pend_event(CA_PEND_EVENT_TIMEOUT);*/
}

/*********************************************************************
  create chid, start search, add connection event handler, add puser,
  add access_rights event handler
 *********************************************************************/
void viewImageCaSearchAndConnectChannel(char *name, chid * pchid,void (*pFunc)())
{
    int status;

    if (strlen(name) <= (size_t) 1) return;
    toBeConnectedCount++;
    status=ca_search_and_connect(name,pchid,viewImageCaConnectionEvent,(void *)pFunc);
    if (status != ECA_NORMAL) {
        errDialog("ca_search_and_connect failed for %s "
            "Return status: %s\n",ca_name(*pchid),
            ca_message(status));
    }
}

/*********************************************************************
  create chid, start search
 *********************************************************************/
void viewImageCaSearchChannel(char *name, chid * pchid)
{
    int status;

    if (strlen(name) <= (size_t) 1) return;
    toBeConnectedCount++;

    status=ca_search(name,pchid);
    if (status != ECA_NORMAL) {
        errDialog("ca_search failed for %s "
            "Return status: %s\n",ca_name(*pchid),
            ca_message(status));
    }
}

/*********************************************************************
 clear a chid
 *********************************************************************/
void viewImageCaClearChannel(chid * pchid)
{
    int status;

    if (!*pchid) return;

    status = ca_clear_channel(*pchid);
    if (status != ECA_NORMAL){
        errDialog("ca_clear_channel failed for %s "
            "Return status: %s\n",ca_name(*pchid),
            ca_message(status));
    }
    *pchid = NULL;
}


/*********************************************************************
 clear event
 *********************************************************************/
void viewImageCaClearEvent(evid * pevid)
{
    int status;

    if (!*pevid) return;

    status = ca_clear_event(*pevid);
    if (status != ECA_NORMAL){
        errDialog("ca_clear_event failed for channel. "
            "Return status: %s\n", ca_message(status));
    }
    *pevid = NULL;
}

/*********************************************************************
 add value event handler
 *********************************************************************/
void viewImageCaAddValueEvent(int request_type, chid chid, evid * pevid,
    void (*pFunc)(struct event_handler_args))
{
    int status;

    if (!chid) return;

    status = ca_add_masked_array_event(request_type, 0,
        chid, pFunc,0,
        (float) 0, (float) 0, (float) 0, pevid, DBE_VALUE);
    if (status != ECA_NORMAL){
        errDialog("ca_add_masked_array_event failed for %s "
        "Return status: %s\n",ca_name(chid),ca_message(status));
    }
}


/*********************************************************************
 Ca sg block
 *********************************************************************/
void viewImageCaSgBlock(const CA_SYNC_GID gid, ca_real timeout)
{
    int status;

    status = ca_sg_block(gid,timeout);
    if (status != ECA_NORMAL){
        errDialog("ca_sg_block failed."
        "Return status: %s\n",ca_message(status));
    }
}

/*********************************************************************
 Ca sg create
 *********************************************************************/
void viewImageCaSgCreate(CA_SYNC_GID * pgid)
{
    int status;

    status = ca_sg_create(pgid);
    if (status != ECA_NORMAL){
        errDialog("ca_sg_create failed."
        "Return status: %s\n",ca_message(status));
         *pgid=0;
    }
}

/*********************************************************************
 Ca sg delete
 *********************************************************************/
void viewImageCaSgDelete(const CA_SYNC_GID gid)
{
    int status;

    status = ca_sg_delete(gid);
    if (status != ECA_NORMAL){
        errDialog("ca_sg_delete failed."
        "Return status: %s\n",ca_message(status));
    }
}

/*********************************************************************
 Ca sg array get
 *********************************************************************/
int viewImageCaSgArrayGet(CA_SYNC_GID gid,chtype type,unsigned long count,chid chid,void *pdata)
{
    int status;

    if (!chid || ca_field_type(chid) == TYPENOTCONN) return -1;
    status = ca_sg_array_get(gid,type,count,chid,pdata);
    if (status != ECA_NORMAL) {
        errDialog("ca_sg_array_get failed for %s "
            "Return status: %s\n",
            ca_name(chid),ca_message(status));
        count=0;
    }
    return count;
}

/*********************************************************************
 Ca array get
 *********************************************************************/
int viewImageCaArrayGet(chtype type,unsigned long count,chid chid,void *pdata)
{
    int status;

    if (!chid || ca_field_type(chid) == TYPENOTCONN) return -1;
    status = ca_array_get(type,count,chid,pdata);
    if (status != ECA_NORMAL) {
        errDialog("ca_array_get failed for %s "
            "Return status: %s\n",
            ca_name(chid),ca_message(status));
        count=0;
    }
    return count;
}


/*********************************************************************
 Ca put
 *********************************************************************/
void viewImageCaPut(int type,chid chid, void *pvalue)
{
    int status;

    if (!chid || ca_field_type(chid) == TYPENOTCONN) return;
    status = ca_put(type, chid, pvalue);
    if (status != ECA_NORMAL) {
        errDialog("ca_put failed for %s "
            "Return status: %s\n",
            ca_name(chid),ca_message(status));
    }
}

#if 0
/*********************************************************************
 Value change event handler
 *********************************************************************/
static void viewImageCaValueEvent(struct event_handler_args args)
{
#if 0
    int stat, sevr;

    stat = ((struct dbr_sts_char *) args.dbr)->status;
    sevr = ((struct dbr_sts_char *) args.dbr)->severity;
    if (sevr>0 ){
        errDialog("viewImageCaValueEvent: Alarm severity=%s: for %s\n",
        alarmSeverityString[sevr],ca_name(args.chid));
    }

    if (stat>0 ){
        errDialog("viewImageCaValueEvent: Alarm status=%s: for %s\n",
        alarmStatusString[stat],ca_name(args.chid));
    }

    switch (args.status) {
    case ECA_NORMAL:
        ((void (*)(struct event_handler_args))args.usr)(args);
        break;
    default:
        errDialog("value event failed: Return status: %s "
        " for %s\n",ca_message(args.status),ca_name(args.chid));
        break;
    }
#endif
}


/*********************************************************************
 Read access event handler
 *********************************************************************/
static void viewImageCaReadAccessEvent(struct access_rights_handler_args args)
{
    if (ca_field_type(args.chid) == TYPENOTCONN) return;
    if (!ca_read_access(args.chid)) {
        errDialog("No read access: %s\n",ca_name(args.chid));
    }
}


/*********************************************************************
 Write access event handler
 *********************************************************************/
static void viewImageCaWriteAccessEvent(struct access_rights_handler_args args)
{
    if (ca_field_type(args.chid) == TYPENOTCONN) return;
    if (!ca_write_access(args.chid)) {
        errDialog("No write access: %s\n",ca_name(args.chid));
    }
}
#endif

/*********************************************************************
 Connection event handler
 *********************************************************************/
static void viewImageCaConnectionEvent(struct connection_handler_args args)
{

    if (args.op == CA_OP_CONN_UP) {
        toBeConnectedCount--;
    } else if (args.op == CA_OP_CONN_DOWN) {
        errDialog("Not Connected: %s\n",
            ca_name(args.chid));
    } else {
        errDialog("Unknown Connection Event: %s\n",
            ca_name(args.chid));
    }
    ((void (*)())ca_puser(args.chid))();
}


/*********************************************************************
 ca exception messages
 *********************************************************************/
static void viewImageCaException(struct exception_handler_args args)
{
#define MAX_EXCEPTIONS 25
    static int nexceptions = 0;
    static int ended = 0;

    if (ended) return;
    if (nexceptions++ > MAX_EXCEPTIONS) {
        ended = 1;
        errDialog("Channel Access Exception: Too many exceptions [%d]\n",
            MAX_EXCEPTIONS);
        ca_add_exception_event(NULL, NULL);
        return;
    }
    errDialog("Channel Access Exception: %s  Context: %s\n",
        ca_message(args.stat) ? ca_message(args.stat) : "Unavailable",
        args.ctx ? args.ctx : "Unavailable");
}


/********************************************************
 Callback when there is activity on a CA file descriptor
********************************************************/
static void viewImageProcessCA(XtPointer cd, int *source, XtInputId *id)
{
    viewImageCaPoll();
}

/********************************************************
 Callback to register file descriptors
********************************************************/
static void registerCA(void *dummy, int fd, int opened)
{
    struct FDLIST *cur,*next;
    int found;

#ifdef WIN32
    intptr_t  viewImageInputMask=XtInputReadWinsock;
#else
    intptr_t  viewImageInputMask=XtInputReadMask;
#endif

    /* Branch depending on whether the fd is opened or closed */
    if(opened) {
        /* Look for a linked list structure for this fd */
        cur=lastFdInList;
        while(cur) {
            if(cur->fd == fd) {
                errDialog("Tried to add a second callback "
                    "for file descriptor %d\n",fd);
                return;
            }
            cur=cur->prev;
        }
        /* Allocate and fill a linked list structure for this fd */
        cur=(struct FDLIST *)calloc(1,sizeof(struct FDLIST));
        if(cur == NULL) {
            errDialog("Could not allocate space to keep track of "
                "file descriptor %d\n",fd);
            return;
        }
        cur->prev=lastFdInList;
        cur->inpid=XtAppAddInput(appContext,fd,(XtPointer)viewImageInputMask,
            viewImageProcessCA,NULL);
        cur->fd=fd;
        lastFdInList=cur;
    } else {
        /* Find the linked list structure for this fd */
        found=0;
        cur=next=lastFdInList;
        while(cur) {
            if(cur->fd == fd) {
                found=1;
                break;
            }
            next=cur;
            cur=cur->prev;
        }
        /* Remove the callback */
        if(found) {
            XtRemoveInput(cur->inpid);
            if(cur == lastFdInList) lastFdInList=cur->prev;
            else next->prev=cur->prev;
            free(cur);
        } else {
            errDialog("Error removing callback for file descriptor %d\n",fd);
        }
    }
}

/*********************************************************************
 cancel timer
 *********************************************************************/
void viewImageCaStop()
{
    /* cancel timeout */
    if (caTimeoutId) {
        XtRemoveTimeOut(caTimeoutId);
        caTimeoutId = (XtIntervalId) 0;
    }
}

/*********************************************************************
 cancel timer, close channel access
 *********************************************************************/
void viewImageCaClose()
{
    int status;

    /* Register file descriptor callback */
    status = ca_add_fd_registration(registerCA, NULL);
    if (status != ECA_NORMAL){
        errDialog("ca_add_fd_registration failed: Return status: %s\n",
        ca_message(status));
    }

    /* close channel access */
    status = ca_task_exit();
    if (status != ECA_NORMAL){
        errDialog("ca_task_exit failed: Return status: %s\n",ca_message(status));
    }
}

