/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
* National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
* Operator of Los Alamos National Laboratory.
* This file is distributed subject to a Software License Agreement found
* in the file LICENSE that is included with this distribution. 
\*************************************************************************/
/* viewImageGui.c */

/* viewImageGui.c,v 1.20 2010/06/23 20:10:46 jba Exp */
/************************DESCRIPTION***********************************
  Image gui routines.
**********************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <inttypes.h>
#include <math.h>

#include <X11/cursorfont.h>
#include <Xm/Xm.h>
#include <Xm/BulletinB.h>
#include <Xm/DrawingA.h>
#include <Xm/Form.h>
#include <Xm/LabelG.h>
#include <Xm/MessageB.h>
#include <Xm/PushBG.h>
#include <Xm/RowColumn.h>
#include <Xm/Scale.h>
#include <Xm/SeparatoG.h>
#include <Xm/SSpinB.h>
#include <Xm/Text.h>
#include <Xm/ToggleB.h>
#include <Xm/ToggleBG.h>

#ifdef USE_SHM
#include <X11/extensions/XShm.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#endif

#include "viewImageErrDialog.h"
#include "viewImageSystemDialog.h"
#include "viewImageData.h"
#include "viewImageGui.h"
#include "viewImageVersion.h"

/* globals */
Display *display;
XtAppContext appContext;
Widget topLevelShell;
Widget drawingArea;

/* # defines */
#define MAX(a,b) ((int)(a)>(int)(b)?(int)(a):(int)(b))
#define MIN(a,b) ((int)(a)<(int)(b)?(int)(a):(int)(b))

#ifndef UNREFERENCED_PARAMETER
#define UNREFERENCED_PARAMETER(x) x = x
#endif

#define MAX_COLORS             256
#define COLOR_SCALE            256

#define FRAME_WIDTH_DEFAULT      512
#define FRAME_HEIGHT_DEFAULT     480

#define WINDOW_MANAGER_WIDTH      10
#define WINDOW_MANAGER_HEIGHT     40

#define POPUP_MENU_UPDATE_IMAGE_DATA_NOW          0
#define POPUP_MENU_UPDATE_IMAGE_DATA_AT_NN_HERTZ  1
#define POPUP_MENU_MODIFY_UPDATE_IMAGE_DATA_HERTZ 2
#define POPUP_MENU_DISPLAY_ROI_ONLY               3
#define POPUP_MENU_DISPLAY_VIEWREGION_ONLY        4
#define POPUP_MENU_SET_ROI_TO_VIEWREGION          5
#define POPUP_MENU_GAIN_CONTROL                   6
#define POPUP_MENU_FIT_TO_IMAGE                   7
#define POPUP_MENU_NEW_SYSTEM                     8
#define POPUP_MENU_PRINT_WINDOW                   9
#define POPUP_MENU_HELP                          10
#define POPUP_MENU_EXIT                          11

#define GRAB_WINDOW None

#define VIDEO_XI_STANDARD     0x00 /* Use standard Xlib calls */
#define VIDEO_XI_SHMSTD       0X01 /* Use Xlib shared memory extension */

/* locals */
static char *systemName=0;
static Dimension xHeight=FRAME_HEIGHT_DEFAULT;
static Dimension xWidth=FRAME_WIDTH_DEFAULT;
static int viewRegionX=0;
static int viewRegionY=0;
static int viewRegionWidth=0;
static int viewRegionHeight=0;

static int gainIsAutomatic = 0;
static double automaticGain = 1.0;
static double scaleGain = 1.0;
static double minGain = 1.0;
static double maxGain = 10.0;

static int dataUpdate=0;
static int dataUpdateRateHertz = 1;
static unsigned long requestImageUpdateInterval=1000;
static unsigned long guiUpdateInterval=100; /* in milliseconds */

static int dataUpdateRateMinHertz = 1;
static int dataUpdateRateMaxHertz = 50;
static int dataUpdateRateIncrementHertz = 1;
static int dataUpdateRateColumns = 2;

static int screenNum;
static Screen *screen;
static Window drawWin=0;
static Window rootWindow;
static Pixel colormap[MAX_COLORS];
static int colorsAlloc=0;
static GC gc;
static GC xorGC;
static XImage  *videoImage=0;
static unsigned int *frameData=0;
static Cursor rubberbandCursor;
static Widget drawingAreaPopupMenu;
static Widget updateNowButton;
static Widget dataUpdateLoopButton;
static Widget modifyDataUpdateRateButton;
static Widget displayRoiOnlyButton;
static Widget displayViewRegionOnlyButton;
static Widget setGainControlButton;
static Widget setRoiButton;
static Widget fitToImageButton;
static Widget newSystemButton;
static Widget printWindowButton;
static Widget helpButton;
static Widget exitButton;
static Widget maxValueLabel = 0;
static Widget minValueLabel = 0;
static Widget gainControlDialog=0;
static Widget gainControlScaleWidget = 0;
static Widget minGainLabel = 0;
static Widget maxGainLabel = 0;
static Widget modifyDataUpdateRateSpinBoxWidget = 0;

static XtIntervalId requestTimeoutId;
static XtIntervalId updateTimeoutId;

#ifdef USE_SHM
static XShmSegmentInfo *shmseginfo = 0;
#endif
static int videoaccesstype=VIDEO_XI_STANDARD;


static String fallbackResources[] = {
    "viewImage*helpDialog*background: LightYellow",
    "viewImage*helpDialog*foreground: Black",
    NULL,
};

static char helpString[]=
"DESCRIPTION:\n"
"    viewImage is an application that displays the images captured by any of the\n"
"    MV200-based imaging systems. It provides the ability to view only a portion\n"
"    of the image, control of the ROI defined on the imaging system, and intensity\n"
"    control for low brightness images.\n"
"\n"
"MENU ENTRIES:\n"
"    Update now\n"
"        Commands the imaging system to grab a new image which is then displayed by\n"
"        viewImage.\n"
"    Update every nn hertz\n"
"        Commands the system to grab a new image at an nn hertz rate, which is then \n"
"        displayed by viewImage. NOTE: This mode may interfere with other applications\n"
"        attempting to grab images from the system. Use this mode only when you are\n"
"        sure no one else is using the system.\n"
"    Modify update rate ...\n"
"        Allows update rate in hertz to be modified. Default rate is 1 Hertz.\n"
"    Display ROI only\n"
"        Expands the image such that only the Region of Interest (as read from \n"
"        the imaging system and displayed with the solid rectangle) is displayed.\n"
"        NOTE: image size will change to maintain the image aspect ratio \n"
"    Display view region only\n"
"        Expands the image such that only the view region (as defined by the user\n"
"        and displayed with the dashed rectangle) is displayed. NOTE: image\n"
"        size will change to maintain the image aspect ratio \n"
"        than the image.\n"
"    Set ROI to view region\n"
"        Changes the Region of Interest setting on the imaging system as defined\n"
"        by the view region.\n"
"    Gain control ...\n"
"        Allows a gain factor to be increased to brighten the intensity of the image.\n"
"        An \"Automatic Gain\" mode will set the gain based on the brightest pixel\n"
"        value in the viewing region.\n"
"    Fit to image\n"
"        Restores the size of the display window to match the image size.\n"
"    New System ...\n"
"        Allows the user to select a different imaging system (e.g. LI:VD1, S:VID1, \n"
"        S:VID2, S:VID3)\n"
"    Help\n"
"        Displays this text.\n"
"    Exit\n"
"        Exits the application.";
/* forward declarations */
static void viewImageGuiDrawingAreaPopupMenuCallback(Widget w,XtPointer calldata,XtPointer cbs);
static void viewImageGuiDrawingAreaCallback(Widget w,XtPointer data,XtPointer cbs);
static void viewImageGuiUpdateImage();
static void viewImageGuiImageDataUpdateLoop(XtPointer client_data,XtIntervalId *id);
static void initColorMap();
static void drawRoiRectangle();
static void drawLocalRoiRectangle();
static void viewImageGuiUpdateXImage();
static  int doRubberbanding(Window window, Position *initialX,
                Position *initialY,Position *finalX,  Position *finalY);
static void initializeRubberbanding(Window window);
static void viewImageGuiSetUpdateNowButtonSensitive();
static void viewImageGuiSetDataUpdateLoopButtonSensitive();
static void viewImageGuiImageUpdateLoop(XtPointer unused,XtIntervalId *id);
static void initDefaultColorMap();
static void viewImageGuiGainControl();
static void viewImagePrintWindow(Widget w);
static void helpDialog(Widget w);
static void viewImageGuiModifyDataUpdateRate();
static void viewImageGuiDisplayViewRegion(int value,
                int subRegionWidth,int subRegionHeight);
static void viewImageGuiExit();
static int getclientbyteorder();


#ifdef USE_SHM
/******************************************************
  set_videoaccestype
******************************************************/
int set_videoaccesstype(Display *dpy, XErrorEvent *ev)
/* a XErrorHandler */
{
  UNREFERENCED_PARAMETER(dpy);
  UNREFERENCED_PARAMETER(ev);
  videoaccesstype=VIDEO_XI_STANDARD;
  return 0;
}
#endif

/******************************************************
  Initialize the application 
******************************************************/
void viewImageGuiInitApp(int *pargc,char *argv[])
{
/* WIN32 initialization */
#ifdef WIN32
  HCLXmInit();
#endif 

  topLevelShell=XtVaAppInitialize(&(appContext),"viewImage",
    NULL,0,pargc,argv,fallbackResources,
    XtNallowShellResize,True,
    NULL);

  display=XtDisplay(topLevelShell);
  screenNum = DefaultScreen(display);
  screen = XtScreen(topLevelShell);
  rootWindow = RootWindow(display,screenNum);
  xorGC = XCreateGC(display,rootWindow,0,NULL);

}

/******************************************************
  Start the main loop 
******************************************************/
void viewImageGuiStartMainLoop()
{
  XtAppMainLoop(appContext);
}

/******************************************************
  Initialize the default colormap 
******************************************************/
static void initDefaultColorMap()
{
  Colormap cmap;
  XColor color,lastColor;
  int i;

  /* X default colormap */
  cmap = DefaultColormap(display,screenNum);

  color.pixel= 0;
  lastColor.pixel= 0;

  if (colorsAlloc) XFreeColors(display,cmap,colormap,MAX_COLORS,0UL);
/* This for loop takes a LONG time from home PC with exceed . */
  for (i = 0; i < MAX_COLORS; i++) {
    /* scale [0,255] to [0,65535] */
    color.red  = (unsigned short) COLOR_SCALE*i;
    color.blue = (unsigned short) COLOR_SCALE*i;
    color.green= (unsigned short) COLOR_SCALE*i;
    /* allocate a shareable color cell with closest RGB value */
    if (XAllocColor(display,cmap,&color)) {
      colormap[i] =  color.pixel;
      lastColor=color;
    } else {
      errDialog("\nmain: couldn't allocate requested color (%d)",i);
      /* put unphysical pixmap value in there as tag it was invalid */
      colormap[i] = lastColor.pixel;
    }
  }
  colorsAlloc=1;
  viewImageDataSetColorTableIsModified(0);
}

/******************************************************
  Initialize the colormap
******************************************************/
static void initColorMap()
{
  Colormap cmap;
  XColor color,lastColor;
  int i;
  unsigned char *redArray;
  unsigned char *blueArray;
  unsigned char *greenArray;

  redArray=viewImageDataGetColorRedArray();
  blueArray=viewImageDataGetColorBlueArray();
  greenArray=viewImageDataGetColorGreenArray();
  if (redArray==0 || blueArray==0 || greenArray==0){
    initDefaultColorMap();
    return;
  }

  color.pixel= 0;
  lastColor.pixel= 0;

  /* X default colormap */
  cmap = DefaultColormap(display,screenNum);
  if (colorsAlloc) XFreeColors(display,cmap,colormap,MAX_COLORS,0UL);

  for (i = 0; i < MAX_COLORS; i++) {
    /* scale [0,255] to [0,65535] */
    color.red  = (unsigned short) COLOR_SCALE*redArray[i];
    color.blue = (unsigned short) COLOR_SCALE*blueArray[i];
    color.green= (unsigned short) COLOR_SCALE*greenArray[i];
    /* allocate a shareable color cell with closest RGB value */
    if (XAllocColor(display,cmap,&color)) {
      colormap[i] =  color.pixel;
      lastColor=color;
    } else {
      errDialog("\nmain: couldn't allocate requested color (%d)",i);
      colormap[i] =  lastColor.pixel;
    }
  }
  colorsAlloc=1;
  viewImageDataSetColorTableIsModified(0);
}

/******************************************************
  viewImageGuiCreateWidgets
******************************************************/
void viewImageGuiCreateWidgets(char *name)
{
  XGCValues gcv;
  Widget separator;
  char       buff[30];
  int       depth;

  /* Create image drawing area */
  drawingArea = XtVaCreateWidget("drawing",
    xmDrawingAreaWidgetClass,topLevelShell,
       NULL);

  /* Create popup menu */
  drawingAreaPopupMenu = XmCreatePopupMenu (drawingArea,
    "popup", NULL, 0);

  /* Create popup menu contents */
  updateNowButton=XtVaCreateManagedWidget("Update now",
        xmPushButtonGadgetClass, drawingAreaPopupMenu,
        NULL);
  XtAddCallback(updateNowButton, XmNactivateCallback,
    viewImageGuiDrawingAreaPopupMenuCallback,(void *)POPUP_MENU_UPDATE_IMAGE_DATA_NOW);

  sprintf(buff,"Update image data at %d Hertz.",dataUpdateRateHertz);
  dataUpdateLoopButton=XtVaCreateManagedWidget(buff,
        xmToggleButtonGadgetClass, drawingAreaPopupMenu,
        NULL);
  XtAddCallback(dataUpdateLoopButton, XmNvalueChangedCallback,
    viewImageGuiDrawingAreaPopupMenuCallback,(void *)POPUP_MENU_UPDATE_IMAGE_DATA_AT_NN_HERTZ);

  modifyDataUpdateRateButton=XtVaCreateManagedWidget("Modify image update rate ...",
        xmPushButtonGadgetClass, drawingAreaPopupMenu,
        NULL);
  XtAddCallback(modifyDataUpdateRateButton, XmNactivateCallback,
    viewImageGuiDrawingAreaPopupMenuCallback,(void *)POPUP_MENU_MODIFY_UPDATE_IMAGE_DATA_HERTZ);

  displayRoiOnlyButton=XtVaCreateManagedWidget("Display ROI only",
        xmToggleButtonGadgetClass, drawingAreaPopupMenu,
        NULL);
  XtAddCallback(displayRoiOnlyButton, XmNvalueChangedCallback,
    viewImageGuiDrawingAreaPopupMenuCallback,(void *)POPUP_MENU_DISPLAY_ROI_ONLY);

  displayViewRegionOnlyButton=XtVaCreateManagedWidget("Display View Region (- - -) only",
        xmToggleButtonGadgetClass, drawingAreaPopupMenu,
        NULL);
  XtAddCallback(displayViewRegionOnlyButton, XmNvalueChangedCallback,
    viewImageGuiDrawingAreaPopupMenuCallback,(void *)POPUP_MENU_DISPLAY_VIEWREGION_ONLY);

  setRoiButton=XtVaCreateManagedWidget("Set ROI to View Region",
        xmPushButtonGadgetClass, drawingAreaPopupMenu,
        NULL);
  XtAddCallback(setRoiButton, XmNactivateCallback,
    viewImageGuiDrawingAreaPopupMenuCallback,(void *)POPUP_MENU_SET_ROI_TO_VIEWREGION);

  setGainControlButton=XtVaCreateManagedWidget("Gain control ...",
        xmPushButtonGadgetClass, drawingAreaPopupMenu,
        NULL);
  XtAddCallback(setGainControlButton, XmNactivateCallback,
    viewImageGuiDrawingAreaPopupMenuCallback,(void *)POPUP_MENU_GAIN_CONTROL);

  fitToImageButton=XtVaCreateManagedWidget("Fit to image",
        xmPushButtonGadgetClass, drawingAreaPopupMenu,
        NULL);
  XtAddCallback(fitToImageButton, XmNactivateCallback,
    viewImageGuiDrawingAreaPopupMenuCallback,(void *)POPUP_MENU_FIT_TO_IMAGE);

  newSystemButton=XtVaCreateManagedWidget("New system",
        xmPushButtonGadgetClass, drawingAreaPopupMenu,
        NULL);
  XtAddCallback(newSystemButton, XmNactivateCallback,
    viewImageGuiDrawingAreaPopupMenuCallback,(void *)POPUP_MENU_NEW_SYSTEM);

  printWindowButton=XtVaCreateManagedWidget("Print window",
        xmPushButtonGadgetClass, drawingAreaPopupMenu,
        NULL);
  XtAddCallback(printWindowButton, XmNactivateCallback,
    viewImageGuiDrawingAreaPopupMenuCallback,(void *)POPUP_MENU_PRINT_WINDOW);

  helpButton=XtVaCreateManagedWidget("Help",
        xmPushButtonGadgetClass, drawingAreaPopupMenu,
        NULL);
  XtAddCallback(helpButton, XmNactivateCallback,
    viewImageGuiDrawingAreaPopupMenuCallback,(void *)POPUP_MENU_HELP);

  separator=XtVaCreateManagedWidget("separator",
        xmSeparatorGadgetClass, drawingAreaPopupMenu,
        NULL);

  exitButton=XtVaCreateManagedWidget("Exit",
        xmPushButtonGadgetClass, drawingAreaPopupMenu,
        NULL);
  XtAddCallback(exitButton, XmNactivateCallback,
    viewImageGuiDrawingAreaPopupMenuCallback,(void *)POPUP_MENU_EXIT);

  /* Add callback for resize events */
  XtAddCallback(drawingArea, XmNresizeCallback,
    viewImageGuiDrawingAreaCallback,NULL);

  /* Add callback for expose events */
  XtAddCallback(drawingArea, XmNexposeCallback,
    viewImageGuiDrawingAreaCallback,NULL);

  /* Add callback for all input events */
  XtAddCallback(drawingArea,XmNinputCallback,
      viewImageGuiDrawingAreaCallback,NULL);

  /* Create a graphic context */
  gcv.foreground = BlackPixelOfScreen(XtScreen(drawingArea));
  gc = XCreateGC(display,rootWindow,GCForeground,&gcv);

  /* Set the System Name */
  viewImageGuiSetSystemName(name);

  XtRealizeWidget(topLevelShell);

  drawWin=XtWindow(drawingArea);

  initializeRubberbanding(rootWindow);

  initColorMap();

  if (!name) systemNameDialog();

#ifdef USE_SHM
{
  int XShmMajor,XShmMinor;
  Bool XShmPixmaps;

  if(XShmQueryVersion(display,&XShmMajor,&XShmMinor,&XShmPixmaps)) {
    videoaccesstype=VIDEO_XI_SHMSTD;
    printf("  # Using Xlib shared memory extension %d.%d\n\n",
             XShmMajor,XShmMinor);
  } else { 
    printf("  # Using conventional Xlib calls.\n\n");
    videoaccesstype=VIDEO_XI_STANDARD;
  }
}
#endif

  depth=DefaultDepth(display,screenNum);

  viewImageGuiUpdateXImage();
  viewImageGuiUpdateImage();

  /* raise error Dialog window if it exists */
  errDialogRaiseWindow();

  viewImageGuiImageUpdateLoop(NULL,0);
}

/******************************************************
  viewImageGuiUpdateXImage
******************************************************/
static void viewImageGuiUpdateXImage()
{


#ifdef USE_SHM
  if(videoaccesstype==VIDEO_XI_SHMSTD) {
    int (*oldhandler)(Display *, XErrorEvent *);

    if (shmseginfo) {
      XShmDetach(display,shmseginfo);
      if(videoImage) XDestroyImage(videoImage);
      shmdt(shmseginfo->shmaddr);
      shmctl(shmseginfo->shmid,IPC_RMID,NULL);
      free(shmseginfo);
    }

    shmseginfo=(XShmSegmentInfo *)malloc(sizeof(XShmSegmentInfo));
    if (!shmseginfo) {
      fprintf(stderr,"  # UpdateXImage: malloc failed for shmseginfo!\n");
      exit(1);
    }
    memset(shmseginfo,0, sizeof(XShmSegmentInfo));

    videoImage = XShmCreateImage(display,
      DefaultVisual(display,screenNum),
      DefaultDepth(display,screenNum),
      ZPixmap,NULL,shmseginfo,
      xWidth,xHeight);
                                   
    if(!videoImage) {
      fprintf(stderr,"  # UpdateXImage: XShmCreateImage failed!\n");
      exit(1);
    }

    /* First create shared memory segment with the size we need, and
       correct permissions for the owner, the group and the world --> 0777 */
    shmseginfo->shmid=shmget(IPC_PRIVATE, videoImage->bytes_per_line*
			         videoImage->height,IPC_CREAT|0777);

    /* Attach the segment to our process, and let the function 
       search the correct memory place --> NULL. It's safest ! */
    shmseginfo->shmaddr=shmat(shmseginfo->shmid,NULL,0);

    /* Set the buffer to Read/Write mode */
    shmseginfo->readOnly=False;

    frameData=(unsigned int *)(videoImage->data = shmseginfo->shmaddr);

    oldhandler = XSetErrorHandler(set_videoaccesstype);
    if(shmseginfo->shmid<0 || !shmseginfo->shmaddr) { 
      videoaccesstype=VIDEO_XI_STANDARD;
    } else {
      if(!XShmAttach(display,shmseginfo)) videoaccesstype=VIDEO_XI_STANDARD;
      XSync(display,0);
    }
    /* if shmget or XShmAttach fails fall back to XCreateImage */
    if (videoaccesstype==VIDEO_XI_STANDARD) {
      shmctl(shmseginfo->shmid, IPC_RMID, 0);
      free(shmseginfo);
      free(videoImage);
      videoImage = NULL;
      /*fprintf(stderr,"%x %p  %d x %d\n",shmseginfo->shmid,shmseginfo->shmaddr,xWidth,xHeight); */
      fprintf(stderr,"  # Shared memory failed, falling back to conventional Xlib calls!\n");
    }

    XSetErrorHandler(oldhandler);
  }
#endif

  if(videoaccesstype==VIDEO_XI_STANDARD) {
    if (videoImage) {
      if (videoImage->data) {
        /* free frameData created by calloc*/
        /* to avoid Exceed problem on WIN32*/
        free(videoImage->data);
        videoImage->data = NULL;
      }
      XDestroyImage(videoImage);
    }
    /* frameData was freed by XDestroyImage */
    frameData = (unsigned int *)calloc(xWidth*xHeight,sizeof(unsigned int));

    videoImage = XCreateImage(display, 
      DefaultVisual(display,screenNum),
      DefaultDepth(display,screenNum),
      ZPixmap,0,(char *)frameData,xWidth,
      xHeight,32,xWidth*sizeof(unsigned int));

      /* set byte order */
      videoImage->byte_order=getclientbyteorder();
  }
}

/******************************************************
  viewImageGuiUpdateImage
******************************************************/
static void viewImageGuiUpdateImage()
{
  int status=0;
  int i,j,k,tmp1,tmp2;
  int maxValue=0,minValue=0,value;
  int imageX,imageY,imageArrayWidth;
  char buff[8];
  int height,width,size,bits;
  float adjustment;
  double temp1,temp2;
  double gain;
  XmString string;
  unsigned char *imageArray=0;

#if 0
time_t now;
now=time(NULL);
/* printf ("viewImageGuiUpdateImage entered  %s",ctime(&now));*/
fflush(stdout);

clock_t start=0,finish;
double duration;
start=clock();
#endif

  if (viewImageDataColorTableIsModified()) initColorMap();

  if (viewImageDataDataIsModified()){
    status=viewImageDataReadImage();
    status=viewImageDataUpdateArrayMinMax();
    maxValue=viewImageDataGetImageMaxValue();
    minValue=viewImageDataGetImageMinValue();

      if (maxValue) {
        automaticGain=((double)(MAX_COLORS-1))/(double)maxValue;
      } else {
        automaticGain=1.0;
      }
      maxGain=10*(((int)(automaticGain/10))+1);

    if (gainControlDialog) {
      if (gainControlScaleWidget) {
        value=maxGain*100;
        XtVaSetValues(gainControlScaleWidget,XmNmaximum,value,NULL);
      }
      if (maxGainLabel) {
        sprintf(buff,"%.2f",maxGain);
        string = XmStringCreateLocalized(buff);
        XtVaSetValues(maxGainLabel,XmNlabelString,string,NULL);
        XmStringFree(string);
      }
      if (maxValueLabel) {
        sprintf(buff,"%d",maxValue);
        string = XmStringCreateLocalized(buff);
        XtVaSetValues(maxValueLabel,XmNlabelString,string,NULL);
        XmStringFree(string);
      }
      if (minValueLabel) {
        sprintf(buff,"%3d",minValue);
        string = XmStringCreateLocalized(buff);
        XtVaSetValues(minValueLabel,XmNlabelString,string,NULL);
        XmStringFree(string);
      }
      if (gainControlScaleWidget) {
        XtVaSetValues(gainControlScaleWidget,XmNvalue,(int)(minGain*100),NULL);
      }
    }
  }


  size=viewImageDataGetImageArraySize();
  bits=viewImageDataGetImageArrayBits();
  if ( bits == 0 ) bits=8*size;

  if (gainIsAutomatic) {
    gain=automaticGain;
    adjustment=1.0;
  } else {
    gain=scaleGain;
    /* Convert range 0 to (2**bits)-1 to range 0 to MAX_COLORS-1 */
    adjustment= (MAX_COLORS-1.0)/(pow(2.0,(float)bits)-1.0);
  }

  width=viewImageDataGetImageWidth();
  height=viewImageDataGetImageHeight();
  imageArray=viewImageDataGetImageArray();
  imageX=viewImageDataGetImageX();
  imageY=viewImageDataGetImageY();
  imageArrayWidth=viewImageDataGetImageArrayWidth();

  temp1=width/(double)xWidth;
  temp2=height/(double)xHeight;

  if (!imageArray) {
    k=MAX_COLORS-1;
    if (k<0) k=0;
    for(i=0; i < xHeight; i++) {
      tmp1 = xWidth*i;
      for(j=0; j < xWidth; j++) {
        frameData[tmp1 + j] = (unsigned int) colormap[k];
      }
    }
  } else {
    if (size==2) {
      for(i=0; i < xHeight; i++) {
        tmp1 = xWidth*i;
        tmp2 = imageArrayWidth*(imageY+(int)(i*temp2));
        for(j=0; j < xWidth; j++) {
          k=(int)(gain*adjustment*(*((unsigned short*)(&imageArray[size*((imageX+(int)(j*temp1))+tmp2)]))));
          if (k>=MAX_COLORS) k = MAX_COLORS - 1;
          if (k<0) k = 0;
          frameData[tmp1 + j] = (unsigned int) colormap[k];
        }
      }
    } else {
      for(i=0; i < xHeight; i++) {
        tmp1 = xWidth*i;
        tmp2 = imageArrayWidth*(imageY+(int)(i*temp2));
        for(j=0; j < xWidth; j++) {
          k=(int)(gain*adjustment*imageArray[(int)(imageX+j*temp1)+tmp2]);
          if (k>=MAX_COLORS) k = MAX_COLORS - 1;
          if (k<0) k = 0;
          frameData[tmp1 + j] = (unsigned int) colormap[k];
        }
      }
    }
  }

#if 0
{
int index=0;
unsigned short val;
if ((xWidth*i+j) < 2000) {
index= ((imageX+(int)(j*temp1))+imageArrayWidth*(int)(imageY+(int)(i*temp2)));
val=(*((unsigned short*)(&imageArray[size*((imageX+(int)(j*temp1)) + imageArrayWidth*(imageY+(int)(i*temp2)))])));
 printf ("size=%d i=%d j=%d imageArray[%d]=%ld gain=%f  k=%d frameData[%d]\n",
size,i,j,index,val,gain,k,(xWidth*i+j));
}
}
#endif


#if 0
finish = clock();
duration = finish-start;
duration = duration/CLOCKS_PER_SEC;
printf("Initialize frameData took %15.10f total sec\n", duration);
fflush(stdout);
start=finish;
#endif

#ifdef USE_SHM
  if(videoaccesstype==VIDEO_XI_SHMSTD) {
    XShmPutImage(display,drawWin,gc,videoImage,0,0,0,0,xWidth,xHeight,False);
  }
#endif
  if(videoaccesstype==VIDEO_XI_STANDARD) {
    XPutImage(display,drawWin,gc,videoImage,0,0,0,0,xWidth,xHeight);
  }
  drawRoiRectangle();
  drawLocalRoiRectangle();
  XFlush(display);

#if 0
finish = clock();
duration = finish-start;
duration = duration/CLOCKS_PER_SEC;
printf ("viewImageGuiUpdateImage took %15.10f total sec\n", duration);
start=finish;
#endif

}

/******************************************************
  viewImageGuiImageUpdateLoop
******************************************************/
static void viewImageGuiImageUpdateLoop(XtPointer unused,XtIntervalId *id)
{
  if(!id && updateTimeoutId) XtRemoveTimeOut(updateTimeoutId);
  updateTimeoutId = 0;
  updateTimeoutId=XtAppAddTimeOut(appContext,guiUpdateInterval,viewImageGuiImageUpdateLoop,0);

  viewImageGuiSetUpdateNowButtonSensitive(viewImageDataUpdateImageIsConnected());
  viewImageGuiSetDataUpdateLoopButtonSensitive(viewImageDataUpdateImageIsConnected());

  if (!id  || viewImageDataColorTableIsModified() || viewImageDataDataIsModified()){

#if 0
{
clock_t start=0,finish;
double duration;
start=clock();
#endif

     viewImageGuiUpdateImage();

#if 0
    viewImageDataReadImage();
finish = clock();
duration = finish-start;
duration = duration/CLOCKS_PER_SEC;
printf ("viewImageGuiUpdateImage took %15.10f total sec\n", duration);
/*printf ("viewImageDataReadImage took %15.10f total sec\n", duration);*/
}
#endif

  }
}

/******************************************************
  viewImageGuiImageDataUpdateLoop
******************************************************/
static void viewImageGuiImageDataUpdateLoop(XtPointer unused,XtIntervalId *id)
{
  if(!id && requestTimeoutId) XtRemoveTimeOut(requestTimeoutId);
  requestTimeoutId = 0;

  if (dataUpdate) {
    requestTimeoutId=XtAppAddTimeOut(appContext,requestImageUpdateInterval,viewImageGuiImageDataUpdateLoop,0);
    viewImageDataRequestImageUpdate();
  }

  if (!id) {
   if (dataUpdate) viewImageGuiSetUpdateNowButtonSensitive(0);
   else viewImageGuiSetUpdateNowButtonSensitive(viewImageDataUpdateImageIsConnected());
  }
}

/******************************************************
  viewImageGuiDrawingAreaPopupMenuCallback
******************************************************/
static void viewImageGuiDrawingAreaPopupMenuCallback(Widget w,
    XtPointer calldata,XtPointer cbs)
{
    int item=(intptr_t)calldata;
    int value;
    int roiX, roiY, roiWidth, roiHeight;
    Dimension width,height;

    switch (item){

    case POPUP_MENU_UPDATE_IMAGE_DATA_NOW:
        viewImageDataRequestImageUpdate();
        break;

    case POPUP_MENU_UPDATE_IMAGE_DATA_AT_NN_HERTZ:
        dataUpdate=(dataUpdate+1)%2;
#if 0   /* GUIUPDATEINTERVAL  */  
        /* set guiUpdateInterval to initial default value */
        if (!dataUpdate) guiUpdateInterval = 100;
#endif   /* GUIUPDATEINTERVAL  */  
        viewImageGuiImageDataUpdateLoop(NULL,NULL);
        break;

    case POPUP_MENU_MODIFY_UPDATE_IMAGE_DATA_HERTZ:
        viewImageGuiModifyDataUpdateRate();
        break;

    case POPUP_MENU_DISPLAY_ROI_ONLY:
        value=XmToggleButtonGetState(w);
        viewImageDataGetRoi(&roiX,&roiY,&roiWidth,&roiHeight);
        if (value){
            if (roiWidth <= 0 || roiHeight <= 0 ) {
              errDialog("Please set a valid ROI first.\n");
              XmToggleButtonSetState(w,0,0);
              break;
            }
            XmToggleButtonSetState(displayViewRegionOnlyButton,0,0);
            viewImageDataSetSubRegion(roiX,roiY,roiWidth,roiHeight);
        }
        viewImageGuiDisplayViewRegion(value,roiWidth,roiHeight);
        break;

    case POPUP_MENU_DISPLAY_VIEWREGION_ONLY:
        value=XmToggleButtonGetState(w);
        if (value){
            if (viewRegionWidth <= 0 || viewRegionHeight <= 0 ) {
              errDialog("Please press mouse button and drag mouse\nto set a valid View Region first.\n");
              XmToggleButtonSetState(w,0,0);
              break;
            }
            XmToggleButtonSetState(displayRoiOnlyButton,0,0);
            viewImageDataSetSubRegion(viewRegionX,viewRegionY,
                viewRegionWidth,viewRegionHeight);
        }
        viewImageGuiDisplayViewRegion(value,viewRegionWidth,viewRegionHeight);
        break;

    case POPUP_MENU_SET_ROI_TO_VIEWREGION:
        if (viewRegionWidth <= 0 || viewRegionHeight <= 0 ) {
          errDialog("Please do mouse button press and drag \nto set a valid View Region first.\n");
        } else {
            viewImageDataSetRoi(viewRegionX,viewRegionY,viewRegionWidth,viewRegionHeight);
            viewRegionWidth=0;
            viewRegionHeight=0;
        }
        break;

    case POPUP_MENU_GAIN_CONTROL:
        viewImageGuiGainControl();
        break;

    case POPUP_MENU_FIT_TO_IMAGE:
        height=viewImageDataGetImageHeight();
        width=viewImageDataGetImageWidth();
        if (height>0 && width>0){
          XtVaSetValues(drawingArea,
            XmNwidth,width,
            XmNheight,height,
            NULL);
          XtManageChild(drawingArea);
        }
        break;

    case POPUP_MENU_NEW_SYSTEM:
        systemNameDialog();
        break;

    case POPUP_MENU_PRINT_WINDOW:
        viewImagePrintWindow(XtParent(w));
        break;

    case POPUP_MENU_HELP:
        helpDialog(w);
        break;

    case POPUP_MENU_EXIT:
        viewImageGuiExit();
        break;
    }
}

/******************************************************
  viewImageGuiExit
******************************************************/
static void viewImageGuiExit()
{
  if (requestTimeoutId) XtRemoveTimeOut(requestTimeoutId);
  if (updateTimeoutId) XtRemoveTimeOut(updateTimeoutId);
  XtDestroyWidget(topLevelShell);
  viewImageDataDisconnect();
  if(videoaccesstype==VIDEO_XI_STANDARD) {
    if (videoImage) {
      if (videoImage->data) {
        /* free frameData created by calloc*/
        /* to avoid Exceed problem on WIN32*/
        free(videoImage->data);
        videoImage->data = NULL;
      }
      XDestroyImage(videoImage);
    }
  }
#ifdef USE_SHM
  if(videoaccesstype==VIDEO_XI_SHMSTD) {
    if (shmseginfo) {
      XShmDetach(display,shmseginfo);
      if(videoImage) XDestroyImage(videoImage);
      shmdt(shmseginfo->shmaddr);
      shmctl(shmseginfo->shmid,IPC_RMID,NULL);
      free(shmseginfo);
    }

  }
#endif
  exit(0);
}

/******************************************************
  viewImageGuiDisplayViewRegion
******************************************************/
static void viewImageGuiDisplayViewRegion(int value,
    int subRegionWidth, int subRegionHeight)
{
  int newHeight, newWidth;
  Position xxWidth,xxHeight;
  Dimension width,height;

  if (value) {
    newWidth=subRegionWidth;
    newHeight=subRegionHeight;
  } else {
    newWidth=viewImageDataGetImageArrayWidth();
    newHeight=viewImageDataGetImageArrayHeight();
  }
  XtVaGetValues(drawingArea, XmNwidth,&xxWidth, NULL);
  XtVaGetValues(drawingArea, XmNheight,&xxHeight, NULL);
  width=xxWidth;
  height=(((double)newHeight)/newWidth)*xxWidth;
  if ( (value && (height>xxHeight)) || (!value && (height<xxHeight)) ) {
    height=xxHeight;
    width=(((double)newWidth)/newHeight)*xxHeight;
  }
  xWidth=width;
  xHeight=height;
  XtVaSetValues(drawingArea,
    XmNwidth,xWidth,
    XmNheight,xHeight,
    NULL);
  XtManageChild(drawingArea);
  viewImageDataSetSubRegionOnly(value);
  viewImageGuiUpdateXImage();
  viewImageGuiUpdateImage();
}

/******************************************************
  viewImageGuiDrawingAreaCallback
******************************************************/
static void viewImageGuiDrawingAreaCallback(Widget w,
    XtPointer client, XtPointer calldata)
{
  XmDrawingAreaCallbackStruct  *cbs;
  XEvent                       *event;
  Dimension                     newHeight,newWidth;
  Position initialX,initialY,finalX,finalY;
  int height,width,xStart,yStart;

  cbs = (XmDrawingAreaCallbackStruct *)calldata;
  event = cbs->event;

  switch (cbs->reason)
  {
      case XmCR_INPUT:
        if (event->xany.type == ButtonPress) {

          if (event->xbutton.button == Button1) {
              XRaiseWindow(display,XtWindow(w));
              XSetInputFocus(display,XtWindow(w),RevertToParent,CurrentTime);
              /* Undraw old rectangle */
              drawLocalRoiRectangle();
              initialX=event->xbutton.x; 
              initialY=event->xbutton.y; 
              finalX=event->xbutton.x; 
              finalY=event->xbutton.y; 
              XSetLineAttributes(display,xorGC,0,LineOnOffDash,CapButt,JoinMiter);
              doRubberbanding(XtWindow(w),&initialX,&initialY,&finalX,&finalY);
              XSetLineAttributes(display,xorGC,0,LineSolid,CapButt,JoinMiter);
#if 0
printf("initial (%d,%d) final (%d,%d) size (%d,%d)\n",initialX,initialY,
finalX,finalY,finalX-initialX,finalY-initialY);
#endif
              height=viewImageDataGetImageHeight();
              width=viewImageDataGetImageWidth();
              xStart=viewImageDataGetImageX();
              yStart=viewImageDataGetImageY();
              viewRegionX=xStart+(initialX*width/xWidth);
              viewRegionY=yStart+(initialY*height/xHeight);
              viewRegionWidth=(finalX-initialX)*width/xWidth;
              viewRegionHeight=(finalY-initialY)*height/xHeight;
              /* Draw new rectangle */
              drawLocalRoiRectangle();

          } else if (event->xbutton.button == Button3) {
              /* popup the menu */
              XmMenuPosition(drawingAreaPopupMenu,
               (XButtonPressedEvent *)event);
              XtManageChild(drawingAreaPopupMenu);
          }
        }
        break;
      case XmCR_RESIZE:
        XtVaGetValues(drawingArea, XmNwidth,&newWidth, NULL);
        XtVaGetValues(drawingArea, XmNheight,&newHeight, NULL);
        if (newWidth!=xWidth || newHeight!=xHeight) {
            width=newWidth;
            height=(((double)xHeight)/xWidth)*newWidth;
            if (height > newHeight ) {
              height=newHeight;
              width=(((double)xWidth)/xHeight)*newHeight;
              
            }
            xWidth=width;
            xHeight=height;
            XtVaSetValues(drawingArea,
              XmNwidth,xWidth,
              XmNheight,xHeight,
              NULL);
            XtManageChild(drawingArea);
            viewImageGuiUpdateXImage();
            viewImageGuiUpdateImage();
        }
        break;

      case XmCR_EXPOSE:
#ifdef USE_SHM
        if(videoaccesstype==VIDEO_XI_SHMSTD) {
          Boolean send_event = False; /* send completion event if True */
          XShmPutImage(event->xany.display,cbs->window,gc,
            videoImage,0,0,0,0,xWidth,xHeight,send_event);
        }
#endif
        if(videoaccesstype==VIDEO_XI_STANDARD) {
          XPutImage(event->xany.display,cbs->window,gc,
            videoImage,0,0,0,0,xWidth,xHeight);
        }
        drawRoiRectangle();
        drawLocalRoiRectangle();
        XFlush(event->xany.display);
        break;
  }
}

/******************************************************
  viewImageGuiGetSystemName
******************************************************/
char * viewImageGuiGetSystemName()
{
   return systemName;
}

/******************************************************
  viewImageGuiSetSystemName
******************************************************/
void viewImageGuiSetSystemName(char *name)
{
  Dimension width,height;
  Position x,y;
  Dimension newWidth,newHeight;
  Dimension maxWidth,maxHeight;
  char appTitle[100];  /* WARNING: Fixed length string */

  if (name && systemName && strcmp(systemName,name)==0) return;

  if (systemName) XtFree(systemName);
  systemName=name;

  dataUpdate=0;
  XmToggleButtonSetState(dataUpdateLoopButton,0,0);
  viewImageGuiImageDataUpdateLoop(NULL,NULL);

  viewImageDataSetSubRegionOnly(FALSE);
  XmToggleButtonSetState(displayRoiOnlyButton,0,0);
  XmToggleButtonSetState(displayViewRegionOnlyButton,0,0);

  viewImageDataCreateData(systemName);

  /* determine drawingArea size */
  newWidth=viewImageDataGetImageArrayWidth();
  newHeight=viewImageDataGetImageArrayHeight();
  if ( newWidth > 0 && newHeight > 0) {
      xWidth=newWidth;
      xHeight=newHeight;
  } else {
    xHeight=FRAME_HEIGHT_DEFAULT;
    xWidth=FRAME_WIDTH_DEFAULT;
  }

  /* reduce drawingArea size to fit on screen */
  maxWidth = WidthOfScreen(screen) - WINDOW_MANAGER_WIDTH;
  maxHeight = HeightOfScreen(screen) - WINDOW_MANAGER_HEIGHT;
  if (maxWidth<xWidth || maxHeight<xHeight) {
      width=maxWidth;
      height=(((double)xHeight)/xWidth)*maxWidth;
      if (height > maxHeight ) {
        height=maxHeight;
        width=(((double)xWidth)/xHeight)*maxHeight;

      }
      xWidth=width;
      xHeight=height;
      errDialog("Image size has been reduced to fit on screen.\n");
  }

  /* set drawingArea size */
  XtVaSetValues(drawingArea,
      XmNwidth,xWidth,
      XmNheight,xHeight,
      NULL);
  XtManageChild(drawingArea);

  /* center widgets on screen */
  x = (maxWidth - xWidth)/2;
  y = (maxHeight - xHeight)/2;
  XtVaSetValues(topLevelShell,XmNx,x,NULL);
  XtVaSetValues(topLevelShell,XmNy,y,NULL);

  if (systemName) {
    sprintf(appTitle,"%s    System  %s",APP_VERSION_STRING,systemName);
  } else {
    sprintf(appTitle,"%s    System  UNSPECIFIED",APP_VERSION_STRING);
  }
  XtVaSetValues(topLevelShell,XmNtitle,appTitle,NULL);

  if (drawWin != 0) {
    viewImageGuiUpdateXImage();
    viewImageGuiUpdateImage();
  }
}

/******************************************************
  drawLocalRoiRectangle
******************************************************/
static void drawLocalRoiRectangle()
{
    int x=0, y=0, width, height, xStart, yStart;
    unsigned int w=0, h=0;

    width=viewImageDataGetImageWidth();
    height=viewImageDataGetImageHeight();
    xStart=viewImageDataGetImageX();
    yStart=viewImageDataGetImageY();
    if (width) {
       x=(viewRegionX-xStart)*xWidth/width;
       w=xWidth*viewRegionWidth/width;
    }
    if (height) {
       y=(viewRegionY-yStart)*xHeight/height;
       h=xHeight*viewRegionHeight/height;
    }
    XSetLineAttributes(display,xorGC,0,LineOnOffDash,CapButt,JoinMiter);
    XDrawRectangle(display,drawWin,xorGC,x,y,w,h);
    XSetLineAttributes(display,xorGC,0,LineSolid,CapButt,JoinMiter);
}

/******************************************************
  drawRoiRectangle
******************************************************/
static void drawRoiRectangle()
{
    int x=0, y=0, width, height, xStart, yStart;
    int roiX, roiY, roiWidth, roiHeight;
    unsigned int w=0, h=0;

    width=viewImageDataGetImageWidth();
    height=viewImageDataGetImageHeight();
    xStart=viewImageDataGetImageX();
    yStart=viewImageDataGetImageY();
    viewImageDataGetRoi(&roiX,&roiY,&roiWidth,&roiHeight);
    if (width) {
       x=(roiX-xStart)*xWidth/width;
       w=xWidth*roiWidth/width;
    }
    if (height) {
       y=(roiY-yStart)*xHeight/height;
       h=xHeight*roiHeight/height;
    }
    XDrawRectangle(display,drawWin,xorGC,x,y,w,h);
}

/******************************************************
  viewImageGuiSetDataUpdateLoopButtonSensitive
******************************************************/
static void viewImageGuiSetDataUpdateLoopButtonSensitive(int val)
{
    static int sensitive=1;

    if (val != sensitive) {
        if (val) XtVaSetValues(dataUpdateLoopButton, XmNsensitive, TRUE, NULL);
        else XtVaSetValues(dataUpdateLoopButton, XmNsensitive, FALSE, NULL);
        sensitive=val;
    }
}

/******************************************************
  viewImageGuiSetUpdateNowButtonSensitive
******************************************************/
static void viewImageGuiSetUpdateNowButtonSensitive(int val)
{
    static int sensitive=1;

    if (val != sensitive) {
        if (val) XtVaSetValues(updateNowButton, XmNsensitive, TRUE, NULL);
        else XtVaSetValues(updateNowButton, XmNsensitive, FALSE, NULL);
        sensitive=val;
    }
}

/******************************************************
  initializeRubberbanding
******************************************************/
static void initializeRubberbanding(Window window)
{
    /*rubberbandCursor = XCreateFontCursor(display,XC_hand2);*/
    rubberbandCursor = XCreateFontCursor(display,XC_crosshair);
    XSetSubwindowMode(display,xorGC,IncludeInferiors);
    XSetFunction(display,xorGC,GXxor);
    XSetForeground(display,xorGC,~0);
}

/******************************************************
  doRubberbanding
  This is code taken from medm and modified
******************************************************/
static int doRubberbanding(Window window, Position *initialX,
  Position *initialY, Position *finalX,  Position *finalY)
{
    XEvent event;
    Boolean returnVal = True;
    int x0, y0, x1, y1;
    static int x=0, y=0;
    static unsigned int w=0, h=0;
    int height,width,xStart,yStart;

    /* Setup */
    *finalX = *initialX;
    *finalY = *initialY;
    x0 = *initialX;
    y0 = *initialY;
    x1 = x0;
    y1 = y0;
    w = (Dimension)0;
    h = (Dimension)0;
    x = MIN(x0,x1);
    y = MIN(y0,y1);

    /* Have ButtonMotionMask|ButtonReleaseMask go to window */
    XGrabPointer(display,window,FALSE,
      (unsigned int)(ButtonMotionMask|ButtonReleaseMask),
      GrabModeAsync,GrabModeAsync,GRAB_WINDOW,rubberbandCursor,CurrentTime);

    /* Grab the server to ensure that XORing will be okay */
    XGrabServer(display);
    XDrawRectangle(display, window, xorGC, x, y, w, h);

    /* Loop until the button is released */
    while(TRUE) {
    XtAppNextEvent(appContext,&event);
    switch (event.type) {
    case ButtonRelease:
      /* Undraw old one */
        XDrawRectangle(display, window, xorGC, x, y, w, h);
        XUngrabServer(display);
        XUngrabPointer(display,CurrentTime);
        XFlush(display);      /* For debugger */
      /* Update current coordinates */
        x1 = event.xbutton.x;
        y1 = event.xbutton.y;
        w =  (MAX(x0,x1) - MIN(x0,x1));
        h =  (MAX(y0,y1) - MIN(y0,y1));
        x = MIN(x0,x1);
        y = MIN(y0,y1);
        *initialX =  x;
        *initialY =  y;
        *finalX   =  x + w;
        *finalY   =  y + h;
        return (returnVal);        /* return from while(TRUE) */

    case MotionNotify:
      /* Undraw old one */
        XDrawRectangle(display, window, xorGC, x, y, w, h);
      /* Update current coordinates */
        x1 = event.xbutton.x;
        y1 = event.xbutton.y;
        w =  (MAX(x0,x1) - MIN(x0,x1));
        h =  (MAX(y0,y1) - MIN(y0,y1));
        x = MIN(x0,x1);
        y = MIN(y0,y1);
      /* Draw new one */
        XDrawRectangle(display, window, xorGC, x, y, w, h);
      /* Update ROI values */
        height=viewImageDataGetImageHeight();
        width=viewImageDataGetImageWidth();
        xStart=viewImageDataGetImageX();
        yStart=viewImageDataGetImageY();
        viewRegionX=xStart+(x*width/xWidth);
        viewRegionY=yStart+(y*height/xHeight);
        viewRegionWidth=w*width/xWidth;
        viewRegionHeight=h*height/xHeight;
        break;
    default:
        XtDispatchEvent(&event);
    }
    }
}

/******************************************************
  viewImageGuiModifyDataUpdateRateCallback
******************************************************/
static void viewImageGuiModifyDataUpdateRateCallback(Widget widget, void *unused, 
  XtPointer callbackStruct)
{
  XmString   string;
  char       buff[30];
  int        reason;
  XmSpinBoxCallbackStruct  *cbs = (XmSpinBoxCallbackStruct *)callbackStruct;

  reason = cbs->reason;
  if (reason != XmCR_OK) return;
  if (dataUpdateRateHertz == cbs->position) return;

  dataUpdateRateHertz = cbs->position;  /* in hertz */
  requestImageUpdateInterval=(long)1000/dataUpdateRateHertz; /* in milliseconds */
  guiUpdateInterval=(long)100/dataUpdateRateHertz; /* in milliseconds */
#if 0   /* GUIUPDATEINTERVAL  */  
  /* set guiUpdateInterval to requestImageUpdateInterval */
  guiUpdateInterval=(long)1000/dataUpdateRateHertz; /* in milliseconds */
#endif   /* GUIUPDATEINTERVAL  */  

  sprintf(buff,"Update image at %d Hertz.",dataUpdateRateHertz);
  string = XmStringCreateLocalized(buff);
  XtVaSetValues(dataUpdateLoopButton, XmNlabelString, string, NULL);
  XmStringFree(string);

}

/******************************************************
  viewImageGuiGainScaleCallback
******************************************************/
static void viewImageGuiGainScaleCallback(Widget widget, void *unused, 
  XtPointer callbackStruct)
{
  XmScaleCallbackStruct *scale_struct=callbackStruct;

  scaleGain = ((double)scale_struct->value)/100.0; 
  viewImageGuiImageUpdateLoop(0,0);
}

/******************************************************
  viewImageGuiAutomaticScaleButtonCallback
******************************************************/
static void viewImageGuiAutomaticScaleButtonCallback(Widget w, XtPointer clientData,
  XtPointer callbackStruct)
{
  int value;

  gainIsAutomatic=(gainIsAutomatic+1)%2;

  if (gainIsAutomatic) {
    XtVaSetValues((Widget)clientData,
      XmNsensitive,FALSE,
      NULL);
  }else {
    value=scaleGain*100;
    XtVaSetValues((Widget)clientData,
      XmNvalue,    value,
      XmNsensitive,TRUE,
      NULL);
  }
  viewImageGuiImageUpdateLoop(0,0);
}

/******************************************************
  viewImageGuiModifyDataUpdateRateDismissCallback
******************************************************/
static void viewImageGuiModifyDataUpdateRateDismissCallback(Widget w, XtPointer clientData,
  XtPointer callbackStruct)
{
  XtUnmanageChild((Widget)clientData);
}

/******************************************************
  viewImageGuiGainDismissCallback
******************************************************/
static void viewImageGuiGainDismissCallback(Widget w, XtPointer clientData,
  XtPointer callbackStruct)
{
  XtUnmanageChild((Widget)clientData);
}

/******************************************************
  viewImageGuiModifyDataUpdateRate
******************************************************/
static void viewImageGuiModifyDataUpdateRate()
{
  static Widget modifyDataUpdateRateDialog=0;
  Widget     formWidget;
  Widget     dismissButton;
  Widget     modifyDataUpdateRateLabel;
  XmString   string;
  int        i;
  Arg        args[20];

  if (modifyDataUpdateRateDialog) {
    XtManageChild(modifyDataUpdateRateDialog);
    XRaiseWindow(display, XtWindow(XtParent(modifyDataUpdateRateDialog)));
    return;
  }

  /* Create Modify Update Rate popup dialog */
  string = XmStringCreateLocalized("viewImage: Modify Update Rate");

  modifyDataUpdateRateDialog = XmCreateBulletinBoardDialog(topLevelShell,"ModifyDataUpdateRate",NULL,0);
    XtVaSetValues(modifyDataUpdateRateDialog,
      XmNdialogTitle, string,
      NULL);

  XmStringFree(string);

  formWidget = XtVaCreateManagedWidget("form", 
    xmFormWidgetClass,modifyDataUpdateRateDialog,
    NULL);

  string = XmStringCreateLocalized("Image update rate in hertz: ");
  modifyDataUpdateRateLabel = XtVaCreateManagedWidget("modifyDataUpdateRateLabel",
    xmLabelGadgetClass, formWidget,
    XmNlabelString,     string,
    XmNtopAttachment,   XmATTACH_FORM,
    XmNtopOffset,       12,
    XmNleftAttachment,  XmATTACH_FORM,
    NULL);
  XmStringFree(string);

  /* Create a numeric SimpleSpinBox */
  i = 0;
  XtSetArg (args[i], XmNspinBoxChildType, XmNUMERIC);                 i++;
  XtSetArg (args[i], XmNminimumValue, dataUpdateRateMinHertz);        i++;
  XtSetArg (args[i], XmNmaximumValue, dataUpdateRateMaxHertz);        i++;
  XtSetArg (args[i], XmNincrementValue, dataUpdateRateIncrementHertz);i++;
  XtSetArg (args[i], XmNeditable, FALSE);                              i++;
  XtSetArg (args[i], XmNcolumns, dataUpdateRateColumns);              i++;
  XtSetArg (args[i], XmNwrap, TRUE);                                 i++;
  XtSetArg (args[i], XmNposition, dataUpdateRateHertz);               i++;
  XtSetArg (args[i], XmNleftAttachment, XmATTACH_WIDGET);             i++;
  XtSetArg (args[i], XmNleftWidget, modifyDataUpdateRateLabel);       i++;

  modifyDataUpdateRateSpinBoxWidget = XmCreateSimpleSpinBox (formWidget,
   "ModifyDataUpdateRate", args, i);

  XtAddCallback(modifyDataUpdateRateSpinBoxWidget, XmNvalueChangedCallback,
    viewImageGuiModifyDataUpdateRateCallback,NULL);

  /* Create Dismiss button */
  dismissButton=XtVaCreateManagedWidget("Dismiss",
    xmPushButtonGadgetClass,formWidget,
    XmNtopAttachment,     XmATTACH_WIDGET,
    XmNtopOffset,         15,
    XmNtopWidget,         modifyDataUpdateRateSpinBoxWidget,
    XmNrightAttachment,  XmATTACH_FORM,
    NULL);

  XtAddCallback(dismissButton,XmNactivateCallback,
    viewImageGuiModifyDataUpdateRateDismissCallback,modifyDataUpdateRateDialog);

  XtManageChild (modifyDataUpdateRateSpinBoxWidget);

  XtManageChild(modifyDataUpdateRateDialog);
}

/******************************************************
  gainControlMenuCallback
******************************************************/
static void viewImageGuiGainControl()
{
  Widget     formWidget;
  Widget     dismissButton;
  Widget     automaticScaleButton;
  Widget     gainLabel;
  Widget     maxValueTextLabel,minValueTextLabel;
  char       buff[8];
  XmString   string;
  int        value;
  int        maxValue,minValue;

  if (gainControlDialog) {
    XtManageChild(gainControlDialog);
    XRaiseWindow(display, XtWindow(XtParent(gainControlDialog)));
    return;
  }

  /* Create Gain Control popup dialog */
  string = XmStringCreateLocalized("viewImage: Gain Control");

  gainControlDialog = XmCreateBulletinBoardDialog(topLevelShell,"Gain Control",NULL,0);
    XtVaSetValues(gainControlDialog,
      XmNdialogTitle, string,
      NULL);

  XmStringFree(string);

  formWidget = XtVaCreateManagedWidget("form", 
    xmFormWidgetClass,gainControlDialog,
    NULL);

  if (gainIsAutomatic) {
    value=automaticGain*100;
  } else {
    value=scaleGain*100;
  }

  /* Create Gain scale */
  gainControlScaleWidget=XtVaCreateManagedWidget("GainControl",
    xmScaleWidgetClass, formWidget,
    XmNmaximum,          (int)(maxGain*100),
    XmNminimum,          (int)(minGain*100),
    XmNdecimalPoints,    2,
    XmNorientation,      XmHORIZONTAL,
    XmNshowValue,        True,
    XmNvalue,            value,
    XmNtopAttachment,    XmATTACH_FORM,
    XmNleftAttachment,   XmATTACH_FORM,
    XmNrightAttachment,  XmATTACH_FORM,
    NULL);

  XtAddCallback(gainControlScaleWidget, XmNvalueChangedCallback,
    viewImageGuiGainScaleCallback,NULL);

  sprintf(buff,"%.2f",minGain);
  string = XmStringCreateLocalized(buff);
 
  minGainLabel = XtVaCreateManagedWidget("minGainLabel",
    xmLabelGadgetClass, formWidget,
    XmNlabelString,     string,
    XmNalignment,       XmALIGNMENT_BEGINNING,
    XmNtopAttachment,   XmATTACH_WIDGET,
    XmNtopWidget,       gainControlScaleWidget,
    XmNleftAttachment,  XmATTACH_FORM,
    NULL);
  XmStringFree(string);

  sprintf(buff,"%.2f",maxGain);
  string = XmStringCreateLocalized(buff);
 
  maxGainLabel = XtVaCreateManagedWidget("maxGainLabel",
    xmLabelGadgetClass, formWidget,
    XmNlabelString,     string,
    XmNalignment,       XmALIGNMENT_END,
    XmNtopAttachment,   XmATTACH_WIDGET,
    XmNtopWidget,       gainControlScaleWidget,
    XmNrightAttachment, XmATTACH_FORM,
    NULL);
  XmStringFree(string);

  string = XmStringCreateLocalized("Gain");
  gainLabel = XtVaCreateManagedWidget("gainLabel",
    xmLabelGadgetClass, formWidget,
    XmNlabelString,     string,
    XmNtopAttachment,   XmATTACH_WIDGET,
    XmNtopWidget,       gainControlScaleWidget,
    XmNrightAttachment, XmATTACH_POSITION,
    XmNrightPosition,   60,
    NULL);
  XmStringFree(string);

  /* Create Automatic Gain toggle button */
  automaticScaleButton=XtVaCreateManagedWidget("Automatic Gain",
    xmToggleButtonGadgetClass, formWidget,
    XmNtopAttachment,   XmATTACH_WIDGET,
    XmNtopWidget,       gainLabel,
    XmNtopOffset,       5,
    XmNleftAttachment,  XmATTACH_FORM,
    XmNrightAttachment, XmATTACH_FORM,
    NULL);

  XtAddCallback(automaticScaleButton, XmNvalueChangedCallback,
    viewImageGuiAutomaticScaleButtonCallback,(void *)gainControlScaleWidget);

  maxValueTextLabel = XtVaCreateManagedWidget("Maximum image value ",
    xmLabelGadgetClass, formWidget,
    XmNtopAttachment,   XmATTACH_WIDGET,
    XmNtopWidget,       automaticScaleButton,
    XmNtopOffset,       15,
    XmNleftAttachment, XmATTACH_FORM,
    NULL);

  maxValue=viewImageDataGetImageMaxValue();
  sprintf(buff,"%3d",maxValue);
  string = XmStringCreateLocalized(buff);

  maxValueLabel = XtVaCreateManagedWidget("maxValueLabel",
    xmLabelGadgetClass, formWidget,
    XmNlabelString,     string,
    XmNtopAttachment,   XmATTACH_WIDGET,
    XmNtopWidget,       automaticScaleButton,
    XmNtopOffset,       15,
    XmNleftAttachment,  XmATTACH_WIDGET,
    XmNleftWidget,      maxValueTextLabel,
    XmNrightAttachment, XmATTACH_FORM,
    NULL);
  XmStringFree(string);

  minValueTextLabel = XtVaCreateManagedWidget("Minimum image value ",
    xmLabelGadgetClass, formWidget,
    XmNtopAttachment,   XmATTACH_WIDGET,
    XmNtopWidget,       maxValueTextLabel,
    XmNleftAttachment, XmATTACH_FORM,
    NULL);

  minValue=viewImageDataGetImageMinValue();
  sprintf(buff,"%3d",minValue);
  string = XmStringCreateLocalized(buff);

  minValueLabel = XtVaCreateManagedWidget("minValueLabel",
    xmLabelGadgetClass, formWidget,
    XmNlabelString,     string,
    XmNtopAttachment,   XmATTACH_WIDGET,
    XmNtopWidget,       maxValueTextLabel,
    XmNleftAttachment,  XmATTACH_WIDGET,
    XmNleftWidget,      minValueTextLabel,
    XmNrightAttachment, XmATTACH_FORM,
    NULL);
  XmStringFree(string);

  /* Create Dismiss button */
  dismissButton=XtVaCreateManagedWidget("Dismiss",
    xmPushButtonGadgetClass,formWidget,
    XmNtopAttachment,     XmATTACH_WIDGET,
    XmNtopOffset,         15,
    XmNtopWidget,         minValueLabel,
    XmNrightAttachment,   XmATTACH_FORM,
    XmNbottomAttachment,  XmATTACH_FORM,
    NULL);

  XtAddCallback(dismissButton,XmNactivateCallback,
    viewImageGuiGainDismissCallback,gainControlDialog);

#if 0
  /* Set Dismiss button as default*/
  XtVaSetValues(formWidget,XmNdefaultButton,dismissButton,NULL);
#endif

  XtManageChild(gainControlDialog);
}

/******************************************************
  Help Dialog 
******************************************************/
static void cancelCallback(Widget w, XtPointer clientdata, XtPointer calldata)
{
    XtUnmanageChild(w);
}

/******************************************************
  Static help Dialog 
******************************************************/
static void helpDialog(Widget w)
{
    static Widget helpDialog=NULL;
    XmString string;
    char dialogName[]="helpDialog"; 
    char dialogTitle[]="viewImage: Help Dialog"; 

  if (helpDialog) {
    XtManageChild(helpDialog);
  } else {
    helpDialog=XmCreateMessageDialog(topLevelShell,dialogName,NULL,0);

    XtVaSetValues(XtParent(helpDialog),
      XmNtitle,dialogTitle,
      XmNbackground,dialogTitle,
      NULL);

    string=XmStringCreateLocalized(helpString);
    XtVaSetValues(helpDialog,XmNmessageString,string,NULL);

    XtUnmanageChild(XmMessageBoxGetChild(helpDialog,XmDIALOG_HELP_BUTTON));
    XtUnmanageChild(XmMessageBoxGetChild(helpDialog,XmDIALOG_CANCEL_BUTTON));
    XtAddCallback(helpDialog,XmNokCallback,cancelCallback,(XtPointer)NULL);
    XtManageChild(helpDialog);
  }
}

/******************************************************
  Static help Dialog 
******************************************************/
static void viewImagePrintWindow(Widget w)
{
    char cmd_buf[256];
    char *psPrinter = 0;
    Window window = 0;

/*NOTE: We may need to stop the window updates */

#if defined(SOLARIS)
    psPrinter = getenv("PSPRINTER");
#endif
    window = XtWindow(topLevelShell);
#if 0
    window = XtWindow(drawingArea);
#endif

    if (psPrinter) {
        sprintf (cmd_buf, "xwd -id %u | xwdtopnm | pnmtops | lp -d%s -onb",
        (unsigned)window, psPrinter);
    }else {
        sprintf (cmd_buf, "xwd -id %u | xwdtopnm | pnmtops | lp -onb",
        (unsigned)window);
    }
    system (cmd_buf);
}

/******************************************************
  Static get byte order of client
******************************************************/
static int getclientbyteorder()
{
    short i=1;

    return (*(char*)&i == 1) ? LSBFirst : MSBFirst;
}

